#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import http.server
import socketserver
from random import choice
from threading import Thread
import os


def get_web_server(port=None):
    '''get_web_server returns a httpd object (socket server)

    Parameters
    ==========
    port: the port for the server, selected randomly if not defined

    '''
    if port == None:
        port = choice(range(2000,9999))

    Handler = http.server.SimpleHTTPRequestHandler
    httpd = socketserver.TCPServer(("", port), Handler)
    server = Thread(target=httpd.serve_forever)
    server.setDaemon(True)
    server.start()
    return httpd,port


def serve_template(webroot, port=None):
    '''serve some folder with a simple httpserver'''

    os.chdir(webroot)
    httpd,port = get_web_server(port=port)
    print('Serving at localhost:%s' %port)
    httpd.serve_forever()



#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__version__ = "0.0.48"
AUTHOR = 'Vanessa Sochat'
AUTHOR_EMAIL = 'vsochat@stanford.edu'
NAME = 'containertree'
PACKAGE_URL = "http://www.github.com/singularityhub/container-tree"
KEYWORDS = 'generate container trees'
DESCRIPTION = "Generate container trees"
LICENSE = "LICENSE"

################################################################################
# Global requirements


INSTALL_REQUIRES = (
    ('requests', {'min_version': '2.18.4'}),
    ('pygments', {'min_version': '2.1.3'}),
)


INSTALL_ANALYSIS = (
    ('pandas', {'min_version': None}),
)

INSTALL_REQUIRES_ALL = (INSTALL_REQUIRES + INSTALL_ANALYSIS)



#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

class Node(object):
    
    def __init__(self, name, attrs, tag=None):
        ''' a Node is a node in the Trie, meaning that
            it stores a word (a folder, name, or file) and some
            number of children from it. If a Node is a leaf 
            this coincides with the end of a path.

            Parameters
            ==========
            name: the name of the folder or file
            attrs: a dict of attributes to give to the node
            tag: a tag or label (goes into a list) to identify objects belonging

        '''
        self.label = name
        self.children = []
        self.set_attributes(attrs)
        
        # The end of the file path
        self.leaf = False

        # If the tag is defined, tag the node
        self.tags = set()
        if tag is not None:
            self.tags.add(tag)

        # How many times this character appeared in the addition process
        self.counter = 1

    def __str__(self):
        return "Node<%s>" % self.label
    def __repr__(self):
        return "Node<%s>" % self.label
    

    def has(tag):
        '''determine if a node has a tag
        '''
        # All supplied tags are in the list
        if isinstance(tag, list):
            if all(tag) in self.tags:
                return True

        # The single tag is in the list
        if tag in self.tags:
            return True
        return False
              

    def get_attributes(self):
        '''return all attributes of the node (aside from children)'''
        ats = {} 
        for key, val in self.__dict__.items():
            if key !="children":
                if isinstance(val, set):
                    val = list(val)
                ats[key] = val
        return ats


    def set_attributes(self, attrs):
        '''Set a variable number of attributes, likely
           size of the file/folder

           Parameters
           ==========
           attrs: a dict of key,value attribute pairs

        '''

        for name, value in attrs.items():
            self.__setattr__(name.lower(), value)

    def get_children(self):
        '''a helper function to get children for a node. This is an iterator,
           if the list is large enough to warrant it.
        '''
        for child in self.children:
            yield child


class MultiNode(Node):
    '''a MultiNode is intended to hold multiple sets of children, indexed by
       the tag. This is intended for a Collection, where each single node
       is a container namespace (e.g., library/ubuntu) and the keys are
       tags, and each tag is associated with a list of children.
    '''
     
    def __init__(self, name, attrs, tag=None):
        super(MultiNode, self).__init__(name, attrs, tag=None)
        self.children = {}

    def __str__(self):
        return "MultiNode<%s>" % self.label
    def __repr__(self):
        return "MultiNode<%s>" % self.label

    def get_children(self):
        '''a helper function to get children for a node. This helps the user
           because the we need to loop over a dictionary. This isn't an iterator
           because it's expected that a node have a reasonably small number
           of children.
        '''
        children = []
        for child_tag in self.children:
            for child in self.children[child_tag]:
                yield child



#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from containertree import ContainerFileTree
from containertree.utils import ( 
    get_tmpdir, 
    get_template, 
    read_json, 
    read_file
)
from containertree.server import serve_template
from containertree.logger import bot
import shutil
import tempfile

import json
import sys
import os


def main(args):
    
    # No image or uri provided, print help
    if len(args.image) == 0:
        subparser.print_help()
        sys.exit(1)

    image = args.image.pop(0)

    # Step 1: Generate container tree object from Docker URI
    tree = ContainerFileTree(image)
    bot.debug(tree)

    # Step 2: Create a webroot to serve from
    webroot = args.output
    if args.output is None:
        webroot = get_tmpdir('containertree-')

    if not os.path.exists(webroot):
        os.mkdir(webroot)

    bot.debug('Webroot: %s' % webroot)

    # Copy the template to the webroot
    template = get_template(args.template)
    shutil.copyfile(template, "%s/index.html" % webroot)

    # If the user wants to print to terminal, we don't save
    filename = None
    if args.printout == None:
        filename='%s/data.json' % webroot

    # Export data.json
    bot.debug('Exporting data for d3 visualization')
    data = tree.export_tree(filename)

    # Does the user want to view the tree?
    if args.view:
        # Generate the data.json
        serve_template(webroot, 9779)

    # Does the user want to print output?
    elif args.printout != None: 
        if args.printout == "data.json":
            print(data)
        elif args.printout == "index.html":
            template = read_file(template)
            print(template)

    return webroot



#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from random import choice
import json
from .node import Node
from .loading import (
    load,
    update,
    _load,
    _update,
    _load_http,
    _load_list,
    _load_json,
    _load_container_diff,
)

class ContainerTreeBase(object):

    def __init__(self, inputs=None, tag=None, folder_sep='/'):
        '''construct a container tree from some export of files or 
           a string to indicate a container. This is determined by
           the subclass.

           Parameters
           ==========
           list: the json export (files.json) of a tar, can be https
           folder_sep: the character that separates path names
           tag: if defined, ascribe a unique id to the list at each node
                to identify the container. We can add other containers to
                the count.
 
        '''

        # The root node is the root of the fs
        self.root = Node('', {'size': 0, 'Name': '/'}, tag=tag)
        self.data = None
        
        # The character that separates folder/files
        self.folder_sep = folder_sep

        # Count the nodes
        self.count = 1
        
        # Sets self.data and builds self.tree
        if inputs != None:
            self.load(inputs)

        # If data is loaded, make the tree
        if self.data:
            self._make_tree(tag=tag)


    def __str__(self):
        return "ContainerTreeBase<%s>" % self.count
    def __repr__(self):
        return "ContainerTreeBase<%s>" % self.count


    def insert(self, name, attrs=None, tag=None):
        '''insert a node into the tree.
        '''
        entry = {'Name': name }
        
        if attrs is not None:
            for key,val in attrs.items():
                entry[key] = val

        self._make_tree(data=[entry], tag=tag)


    def add(self, name, node, attrs={}, tag=None):
        '''add a node based on name to the tree, or return found node
        '''             
        # Do we have the package?    
        found = False
        for child in node.children:

           # We found the parent
           if child.label == name:

               # Keep track of how many we have
               child.counter += 1

               # update node to be child that was found
               node = child
               found = True
               break

        # If we get down here, not found, add new node
        if not found:
            new_node = Node(name, attrs)
            self.count +=1

            # Add to the root (or the last where found)
            node.children.append(new_node)
            node = new_node

        # Add the tag to the new (or existing) node
        if tag is not None:
            if tag not in node.tags:
                node.tags.add(tag)

        return node


    def _make_tree(self, data=None, tag=None):
        '''must be instantiated by subclass.'''
        print('_make_tree must be instantiated by the subclass.')


    def export_tree(self, filename=None):
        '''export a data structure for a weighted, colored tree, either just
           the data or the full html / visualization. If filename is defined,
           export the raw data json to the file. If the generate_html is
           True, do the same but embed in html.

           Parameters
           ==========
           filename: if defined, write data to file (and return) otherwise
           return data structure

        '''

        # We will call this recursively
 
        def traverse(nodes={}, current=None):

            colors = ['#0000FF', # blue
                      '#FF7F00', # orange
                      '#FF0000', # red
                      '#7F007F', # purple
                      '#00FFFF', # cyan
                      '#0560D0'] # generic blue

            if current is None:
                current = self.root

            tags = list(current.tags)
            new_node = {'color': choice(colors),
                        'key': current.label,
                        'name': current.label.split('/')[-1],
                        'tags': tags,
                        'attrs': current.get_attributes(),
                        'children': [] }

            # Add the size if was provided!
            if hasattr(current, 'size'):
                new_node['size'] = current.size

            if len(nodes) == 0:
                nodes.update(new_node)
            else:            
                nodes['children'].append(new_node)

            # Iterate through children, add to data structure
            for child in current.children:
                traverse(nodes=new_node, current=child)

        nodes = dict()

        if self.data not in [None, [], {}, ""]:
            traverse(nodes)
                
        # If the user provided a file, export to it
        if filename is not None:
            with open(filename, 'w') as filey:
                filey.writelines(json.dumps(nodes))
            return filename

        return nodes


# Searching Functions

    def similarity_score(self, tags):
        '''calculate a similarity score for one or more tags. The score is
           a basic information coefficient where we take into account:
 
           1/ the number of total nodes in context of the tags, meaning
              we add any node that has one or more of containers there
           2/ the number of nodes where the tags are all present
           3/ the number of nodes where one or more tags are missing

        '''
        total = 0       # total number of nodes with one or more
        intersect = 0   # all tags present at nodes
        diff = 0        # one or more tags missing

        def traverse(tags, current, total, intersect, diff):
 
            # All tags are represented in the node
            if all(t in current.tags for t in tags):
                intersect+=1
            else:
                diff+=1

            # If any of the tags are present, we add to total
            if any(t in current.tags for t in tags):
                total+=1

            # Iterate through children, add to data structure
            for child in current.children:
                total,intersect,diff = traverse(tags, child, total, intersect, diff)

            return total, intersect, diff

        # Return data structure so the user knows the components
        total, intersect, diff = traverse(tags, self.root, total, intersect, diff)
 
        result = {'total': total, 
                  'tags': tags,
                  'same': intersect,
                  'diff': diff }

        # Calculate score, percentage of nodes shared

        result['score'] = 0
        if total > 0:
           result['score'] = intersect / total

        return result


    def trace(self, name, node=None):
        '''find a path in the tree and return the node if found.
           This base function is suited for searches that don't build
           on themselves (e.g., not filepaths or words)
         '''
        if node == None:
            node = self.root

        # Find the node, is it in the tree?
        tracedNode = self.find(name)

        # Trace it's path
        if tracedNode != None:
            for child in node.children:
                traces = self._trace(tracedNode, child)
                if tracedNode in traces:
                    return [self.root] + traces


    def _trace(self, tracedNode, node, traces=None):
        '''find a path in the tree and return the node if found.
           This base function is suited for searches that don't build
           on themselves (e.g., not filepaths or words)

           Parameters
           ==========
           tracedNode: the node we are looking for.
           node: the current node we are searching
           traces: a list of nodes we have traced thus far
        '''
        if traces == None:
            traces = []

        # Always add node
        traces.append(node)

        # Did we find the node?
        if tracedNode in node.children:
            return traces + [tracedNode]

        # Keep searching over children        
        for child in node.children:
            self._trace(tracedNode, child, traces)
        return traces
       
    def get_count(self, name):
        '''find a path in the tree and return the node if found
        '''
        counter = 0
        node = self.find(name)
        if node:
            counter = node.counter
        return counter


    def find(self, name, node=None):
        '''find a path in the tree and return the node if found.
           This base function is suited for searches that don't build
           on themselves (e.g., not filepaths or words)
         '''

        if node == None:
            node = self.root

        # Did we find a node?
        if node.label == name:
            return node

        # No children, we finished search
        if len(node.children) == 0:
            return None

        for child in node.children:
            node = self.find(name, child)
            if node != None:
                return node


    def remove(self, name, node=None):
        '''find a path in the tree and remove (and return) the node if found.
           The function returns None if the Node wasn't in the tree (and wasn't
           found and removed). The way this is designed, we cannot remove the
           root node (but we just return it).
         '''

        if node == None:
            node = self.root

        # Did we find the node?
        if node.label == name:
            return node

        # No children, we finished search
        if len(node.children) > 0:
            for c in range(len(node.children)):
                child = node.children[c]
                to_remove = self.remove(name, child)
                if to_remove != None:
                    del node.children[c]
                    return to_remove


    def search(self, name, number=None, node=None):
        '''find a basename in the tree. If number is defined, return
           up to that number. 
        '''
        found = []
 
        # We always start at the root  
        if node == None:
            node = self.root

        # Look for the name in the current node
        if name in node.label:
            found.append(node)

        # No children, no search
        if len(node.children) == 0:
            return found

        # Does the user want to cut out early?
        if number != None:
            if len(found) >= number:
                return found

        # Not at current node, search children!
        for child in node.children:
            found += self.search(name, number, child)
            if number != None:
                if len(found) >= number:
                    return found
        return found


# Loading Functions
ContainerTreeBase.load = load
ContainerTreeBase._load = _load
ContainerTreeBase.update = update
ContainerTreeBase._update = _update
ContainerTreeBase._load_http = _load_http
ContainerTreeBase._load_json = _load_json
ContainerTreeBase._load_list = _load_list
ContainerTreeBase._load_container_diff = _load_container_diff



#!/usr/bin/env python
#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from random import choice
from containertree.utils import check_install
import requests
import json
import os
import re
import sys

from .base import ( ContainerTreeBase, Node )


class ContainerTree(ContainerTreeBase):

    def __init__(self, inputs=None, tag=None, folder_sep="/"):
        super(ContainerTree, self).__init__(inputs, tag, folder_sep)

    def __str__(self):
        return "ContainerTree<%s>" % self.count
    def __repr__(self):
        return "ContainerTree<%s>" % self.count

    def _make_tree(self, data=None, tag=None):
        '''construct the tree from the loaded data (self.data)
           we should already have a root defined. Since we are making
           the tree for an individual container, the node names are filepaths
           within the container.
        '''

        # If function is used for insert, called
        if data is None:
            data = self.data

        for attrs in data:

            # The starting node is the root node
            node = self.root

            # Add the tag to the new (or existing) node
            if tag is not None:
                if tag not in node.tags:
                    node.tags.add(tag)

            filepaths = attrs['Name'].split(self.folder_sep)
        
            # Add the path to the correct spot in the tree    
            for filepath in filepaths:

                if not filepath:
                    continue

                found = False

                # We are at the root
                if filepath == node.label:
                    found = True             
                    
                else:

                    # Search in present node
                    for child in node.children:

                        # We found the parent
                        if child.label == filepath:

                            # Keep track of how many we have
                            child.counter += 1

                            # update node to be child that was found
                            node = child
                            found = True
                            break


                # If not found, add new node (child)
                if not found:
                    new_node = Node(filepath, attrs)
                    self.count +=1

                    # Add to the root (or the last where found)
                    node.children.append(new_node)

                    # Keep working down the tree
                    node = new_node

                # Add the tag to the new (or existing) node
                if tag is not None:
                    if tag not in node.tags:
                        node.tags.add(tag)

            # The last in the list is the leaf (file)
            node.leaf = True


class ContainerDiffTree(ContainerTree):
    '''a container diff tree is a subclass of ContainerTree, specifically
       ready to read in an analysis result of a files export from Google's 
       Container Diff. We simply define the _load method to expect the format of:

       [0]['Analysis'] --> [{"Name":"...", "Size": 123 }]

    '''
    def _load(self, data=None):
        return self._filter_container_diff(data, analyze_type="File")

    def _filter_container_diff(self, data=None, analyze_type="File"):
        ''' class instantiated by subclass to do custom parsing of loaded data.
            In the case of Google container diff, whether from local file
            or web http, we find the "File" Analysis type and return it.

            Parameters
            ==========
            data: if provided, use instead of self.data already with instance
            analyze_type: the key to use for the index of Container-Diff
                          can be one of File, Apt, or Pip
        '''
        if data is None:
            data = self.data

        if not data:
            print('This function should be called with load() to define data.')
            sys.exit(1)

        # User can provide loaded data, as long as correct structure
        if not isinstance(data, list):
            print('Loaded Filelist must be list for Container Diff')

        else:
            for entry in data:
                if "Analysis" in entry and "AnalyzeType" in entry:
                    if entry['AnalyzeType'] == analyze_type:
                        analysis = entry['Analysis']

                        # Warn the user if empty
                        if analysis in [[],{},'', None]:
                            print('Warning, no data found for %s' % analyze_type)
                        return analysis

            # If we get down here, tell the user cannot find what looking for
            if len(data) == 0:
                print('Loaded Container Diff Analysis List is empty')

            else:
                print('%s key missing, is this ContainerDiff export?' % analyze_type)


class ContainerFileTree(ContainerDiffTree):
    '''a container file tree will build a file hierarchy tree using the 
       Container Diff File export. This is the base implementation of 
       ContainerTree, so we don't need to write a function to generate
       the tree here.
    '''

    def _load(self, data=None):
        return self._filter_container_diff(data, analyze_type="File")

    def find(self, filepath):
        '''find a path in the tree and return the node if found. For a Container
           Tree, we expect a filepath. This must be the absolute filepath.
           To do a search, use search instead.
        '''

        # We always start at the root  
        node = self.root

        # No children, no search
        if len(node.children) == 0:
            # They were looking for the root
            if node.label == filepath:
                return node
            return
        
        filepaths = filepath.split(self.folder_sep)
        assembled='/'.join(filepaths)

        # Look for the path in the tree
        for filepath in filepaths:

            if not filepath:
                continue

            # Comparing basenames (<etc>)
            if filepath == node.label:

                # Comparing complete assembled path, return node
                if filepath == assembled:
                    return node

            else:

                # Try and find the filepath in the tree
                for child in node.children:
                    if child.label == filepath:
                        node = child

                        # If the name is what we are looking for, return Node
                        if node.name == assembled:
                            return node
                        break


class ContainerPackageTree(ContainerDiffTree):
    '''a container package tree will generate a container tree based on some
       package manager. Since the output is from container-diff, we expect
       the loaded data to conform to the following, a list of:

       [{'Name': 'zlib1g', 'Size': 159744, 'Version': '1:1.2.8.dfsg-5'}..] 
    '''

    def __str__(self):
        return "Container%sTree<%s>" % (self.analyze_type, self.count)
    def __repr__(self):
        return "Container%sTree<%s>" % (self.analyze_type, self.count)


    def export_vectors(self, node=None, df=None, include_tags=None, 
                             skip_tags=None, regexp_tags=None, 
                             include_versions=False, level=0,
                             package=None):
        '''export one or more vectors to describe containers mapped to a
           package tree. Optionally, the user can include or disclude a set
           of containers, or include/disclude version strings.

           Parameters
           ==========
           node: the node to start the export at (default is root)
           df: the pandas dataframe to update or continue adding to
           include_tags: a list of container uris to include
           skip_tags: a list o container uris to skip
           regexp_tags: include tags based ona regular expression
           include_versions: optionally include versions as part of the features
           package: if the user wants versions, we pass the parent package to
                    the child version node
        '''
        # Only import pandas once
        if "pandas" not in locals():
            import pandas

        # Checks if we have initialized the df yet (can't compare to None)
        if not hasattr(df, 'add'):
            df = pandas.DataFrame()

        if node == None:
            node = self.root

        # Skip the root node (label is '')
        if node.label != '':
            
            containers = node.tags

            # If the user is limiting the containers to include
            if include_tags != None:
                containers = containers.intersection(set(include_tags))

            # If the user is skipping containers or tags
            if skip_tags != None:
                containers = [x for x in containers if x not in skip_tags]

            # If the user wants regular expression filtering
            if regexp_tags != None:
                containers = [x for x in containers if re.search(regexp_tags,x)]

            label = node.label
            add_packages = False

            # If the user wants to include versions, and we are on the verison level
            if include_versions and level == 2:
                label = '%s-v%s' %(package, label)
                add_packages = True

            elif not include_versions and level == 1:
                add_packages = True

            # Add the node packages to the tree
            if add_packages:
                for container in containers:
                    df.loc[container, label] = 1

        # Go through children
        level += 1
        for child in node.children:
            df = self.export_vectors(child, df, include_tags, 
                                                skip_tags, 
                                                regexp_tags,
                                                include_versions,
                                                level,
                                                node.label)
        
        return df


    def _make_tree(self, data=None, tag=None):
        '''construct the tree from the loaded data (self.data)
           we should already have a root defined. Since we are making
           the tree for an individual container, the node names are filepaths
           within the container.
        '''

        # If function is used for insert, called
        if data is None:
            data = self.data

        for package in data:

            # The starting node is the root node
            node = self.root

            # Add the tag to the new (or existing) node
            if tag is not None:
                if tag not in node.tags:
                    node.tags.add(tag)
        
            # Add the new node to the tree based on package name
            new_node = self.add(package['Name'], node, tag=tag)

            # Add the version Node
            version_node = self.add(package['Version'], new_node, tag=tag)   

            # The last in the list is the leaf (file)
            version_node.leaf = True


class ContainerPipTree(ContainerPackageTree):
    '''a container pip tree will generate a container tree based on pip
       packages.
    '''
    def __init__(self, inputs=None, tag=None, folder_sep="/"):
        self.analyze_type = "Pip"
        super(ContainerPipTree, self).__init__(inputs, tag, folder_sep)

    def _load(self, data=None):
        return self._filter_container_diff(data, self.analyze_type)


class ContainerAptTree(ContainerPackageTree):
    '''a container apt tree will generate a container tree based on pip
       packages.
    '''
    def __init__(self, inputs=None, tag=None, folder_sep="/"):
        self.analyze_type = "Apt"
        super(ContainerAptTree, self).__init__(inputs, tag, folder_sep)

    def _load(self, data=None):
        return self._filter_container_diff(data, self.analyze_type)



#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
from subprocess import (
    Popen,
    PIPE,
    STDOUT
)
import fnmatch
import os
from glob import glob
import tempfile

def get_installdir():
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def recursive_find(base, pattern=None):
    '''recursively find files that match a pattern, in this case, we will use
       to find Dockerfiles

       Paramters
       =========
       base: the root directory to start the seartch
       pattern: the pattern to search for using fnmatch
    '''
    if pattern is None:
        pattern = "*"

    for root, dirnames, filenames in os.walk(base):
        for filename in fnmatch.filter(filenames, pattern):
            yield os.path.join(root, filename)


def read_json(filename, mode='r'):
    '''read_json reads in a json file and returns
       the data structure as dict.
    '''
    with open(filename, mode) as filey:
        data = json.load(filey)
    return data


def read_file(filename, mode='r'):
    '''read_text file and return string of content
    '''
    with open(filename, mode) as filey:
        content = filey.read()
    return content


def get_template(name):
    '''return an html template based on name.
    '''
    here = get_installdir()
    template_folder = os.path.join(here, 'templates')
    template_file = "%s/%s.html" %(template_folder, name)
    if os.path.exists(template_file):
        return template_file


def get_templates():
    '''list ids of html templates (based on name, without extension).
    '''
    here = get_installdir()
    template_folder = os.path.join(here, 'templates')
    templates = glob('%s/*.html' % template_folder)
    templates = [os.path.basename(x).replace('.html', '') for x in templates]
    return templates


def get_tmpfile(prefix="", tmpdir=None):
    '''get a temporary file with an optional prefix.

       Parameters
       ==========
       prefix: prefix the file with this string.
       tmpdir: the root directory to use (defaults to new temporary folder)
    '''
    if tmpdir == None:
        tmpdir = tempfile.mkdtemp()
    prefix = os.path.join(tmpdir, os.path.basename(prefix))
    fd, tmp_file = tempfile.mkstemp(prefix=prefix)
    os.close(fd)
    return tmp_file


def get_tmpdir(prefix=""):
    '''get a temporary folder with an optional prefix.

       Parameters
       ==========
       prefix: prefix the file with this string.
    '''
    tmpdir = tempfile.gettempdir()
    tmpfile = get_tmpfile(prefix, tmpdir)
    os.remove(tmpfile)    
    os.mkdir(tmpfile)
    return tmpfile


def run_command(cmd, sudo=False):
    '''run_command uses subprocess to send a command to the terminal.

        Parameters
        ==========
        cmd: the command to send, should be a list for subprocess
        error_message: the error message to give to user if fails,
        if none specified, will alert that command failed.

    '''
    if sudo is True:
        cmd = ['sudo'] + cmd

    try:
        output = Popen(cmd, stderr=STDOUT, stdout=PIPE)

    except FileNotFoundError:
        cmd.pop(0)
        output = Popen(cmd, stderr=STDOUT, stdout=PIPE)

    t = output.communicate()[0],output.returncode
    output = {'message':t[0],
              'return_code':t[1]}

    if isinstance(output['message'], bytes):
        output['message'] = output['message'].decode('utf-8')

    return output



def check_install(software='container-diff', quiet=True):
    '''check_install to ensure we have container-diff installed
       before trying to use it.

       Parameters
       ==========
       software: the software to check if installed
       quiet: should we be quiet? (default True)
    '''
    cmd = ['which', software]
    try:
        found = run_command(cmd)
    except: # FileNotFoundError
        return False

    if found['return_code'] == 0:
        if quiet is False:
            print(found['message'])
        return True
    return False



#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# Loading data helper functions are consistent between trees, and shared here.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from containertree.logger import bot
from containertree.utils import ( 
    check_install, 
    run_command,
    read_json,
    get_tmpfile
)
import os
import re
import sys
import json
import requests

def _update(self, inputs, tag=None):
    '''_update is a helper function for update and load. We return
       data based on the inputs provided, and return loaded to the
       calling function. The subsequent action is up to the calling
       function.

       Parameters
       ==========
       inputs: the list of files (json/url export from ContainerDiff,
                 OR the uri of a container (to run container-diff).
       tag: if defined, a tag or label to identify
    '''
    data = None

    # Load data from web / url
    if re.search("https?://", inputs):
        data = self._load_http(inputs)

    # Load data from file
    elif os.path.exists(inputs):
        if inputs.endswith('json'):
            data = self._load_json(inputs)

        # Otherwise, pass on the filepath to the _load function
        else:
            print('Unrecognized extension, passing %s to _load subclass' % inputs) 
            data = inputs

    # Last effort is to run container-diff
    elif check_install(quiet=True):
        data = self._load_container_diff(inputs)
        if not data:
            bot.warning('No container-diff output found for %s' % inputs)
    else:
         print('Error loading %s' % inputs)
    return data


def update(self, inputs, tag=None):
    '''update will load in new data (without distributing an old self.data)

       Parameters
       ==========
       inputs: the list of files (json/url export from ContainerDiff,
                     OR the uri of a container (to run container-diff).
       tag: if defined, a tag or label to identify
    '''
    data = self._update(inputs)

    # If we have loaded data, continue
    if data:
        data = self._load(data)
        self._make_tree(data=data, tag=tag)


def load(self, inputs):
    ''' Load a set of files from json into the container tree.
        This means:
 
        1/ Reading in the data file
        2/ Creating a Node for each path portion
        3/ Assembling the nodes into the tree

        The data format must be a list of files, minimally
        with "Name" corresponding to the full path.

    '''
    data = self._update(inputs)
    if data not in [None, [], {}]:
        self.data = data
        self.data = self._load()

def _load(self, data=None):
    '''a function called by load, intended for subclass to call if additional
       parsing is needed.
    '''
    return self.data


def _load_http(self, url):
    '''load json from http. We assume it to be json because other formats
       aren't supported yet
    '''
    response = requests.get(url)

    if response.status_code == 200:
        try:
            return response.json()
        except json.JSONDecodeError:
            print('JsonDecodeError of web address!')
            sys.exit(1)

def _load_json(self, inputs):
    '''read the inputs from a json file
    '''
    # Read in the raw data file
    with open(inputs) as filey:
        data = json.load(filey)
    return data


def _load_list(self, inputs):
    '''load a inputs. If the files are found to exist and Size is not
       included, calculate it.
    '''
    finished = []
    for entry in inputs:

        # If it's a filepath, convert to dictionary
        if not isinstance(entry, dict):
            entry = {'Name': entry}

        if "Name" not in entry:
            print('Skipping %s, no "Name" provided!' %entry)
            continue

        # If we don't have a size and the file exists, calculate one
        if "Size" not in entry and os.path.exists(entry['Name']):
            entry['Size'] = os.path.getsize(entry['Name'])
        finished.append(entry) 

    return finished


def _load_container_diff(self, container_name, output_file=None, types=None):
    '''call container-diff directly on the command line to extract
       the layers of interest.
    '''
    layers = dict()

    # Stop short if we don't have container-diff
    if not check_install(quiet=True):
        print('container-diff executable not found, cannot extract %s' % inputs)
        return layers

    if types == None:
        types = ['pip', 'apt', 'history', 'file']

    # Common error to provide just a string type
    if not isinstance(types, list):
        types = [types]

    types = ["--type=%s" % t for t in types]

    if output_file == None:
        output_file = get_tmpfile(prefix="container-diff")

    cmd = ["container-diff", "analyze", container_name]
    response = run_command(cmd + types + ["--output", output_file, "--json",
                                          "--quiet", "--no-cache",
                                          "--verbosity=panic"])

    if response['return_code'] == 0 and os.path.exists(output_file):
        layers = read_json(output_file)
        os.remove(output_file)
    else:
        print(response['message'])

    return layers



#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .https import ( get, call )
from .fileio import (
    get_installdir,
    read_json,
    get_template,
    get_templates,
    get_tmpfile,
    get_tmpdir,
    read_file,
    recursive_find,
    run_command,
    check_install    
)

from .docker import (
    parse_image_uri,
    DockerInspector
)



#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import requests

def call(self, url, func, data=None,
                          headers=None, 
                          return_json=True,
                          retry=True,
                          default_headers=True,
                          quiet=False):

    '''call will issue the call, and issue a refresh token
       given a 401 response, and if the client has a _update_token function

       Parameters
       ==========
       func: the function (eg, post, get) to call
       url: the url to send file to
       headers: if not None, update the client self.headers with dictionary
       data: additional data to add to the request
       return_json: return json if successful
       default_headers: use the client's self.headers (default True)

    '''
 
    if data is not None:
        if not isinstance(data, dict):
            data = json.dumps(data)

    heads = dict()
    if default_headers is True:
        heads = self.headers.copy()
    
    if headers is not None:
        if isinstance(headers, dict):
            heads.update(headers)

    response = func(url=url,
                    headers=heads,
                    data=data)

    # Errored response, try again with refresh
    if response.status_code in [500, 502]:
        print("Beep boop! %s: %s" %(response.reason,
                                    response.status_code))
        sys.exit(1)

    # Errored response, try again with refresh
    if response.status_code == 404:

        # Not found, we might want to continue on
        if quiet is False:
            prints("Beep boop! %s: %s" %(response.reason,
                                         response.status_code))
        sys.exit(1)


    # Errored response, try again with refresh
    if response.status_code == 401:

        # If client has method to update token, try it once
        if retry is True and hasattr(self,'update_token'):

            # A result of None indicates no update to the call
            self.update_token(response)
            return self._call(url, func, data=data,
                              headers=headers,
                              return_json=return_json,
                              stream=stream, retry=False)

        bot.error("Your credentials are expired! %s: %s" %(response.reason,
                                                           response.status_code))
        sys.exit(1)

    elif response.status_code == 200:

        if return_json:

            try:
                response = response.json()
            except ValueError:
                bot.error("The server returned a malformed response.")
                sys.exit(1)

    return response

def get(self,url,
             headers=None,
             token=None,
             data=None,
             return_json=True,
             default_headers=True,
             quiet=False):

    '''get will use requests to get a particular url
    '''
    print("GET %s" % url)
    return self._call(url,
                      headers=headers,
                      func=requests.get,
                      data=data,
                      return_json=return_json,
                      default_headers=default_headers,
                      quiet=quiet)



#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from copy import deepcopy
from random import choice
from containertree.utils import (
    check_install, 
    parse_image_uri
)
import requests
import json
import re

from .node import ( MultiNode, Node )
from .loading import (
    _load_http,
    _load_list,
    _load_json,
    _load_container_diff,
    _update
)

class CollectionTree(object):

    def __init__(self, inputs=None, tag=None, first_level='library'):
        '''a collection tree is similar to a container tree, except instead
           of a filesystem we have a tree of containers, where we walk up
           the tree (finding parents via the Dockerfile) until we get to scratch
           or cannot proceed (with a 404, etc.) We produce a tree of containers
           that are based on FROM statements, and we obtain the metadata about
           the containers from container-diff. The root of the tree (scratch)
           is the only one to be of type "Node." We define the base family
           of the tree as library, and this is used to control what can be 
           added. If you don't want to limit the root node's first level,
           then set first_level to an empty string.
        '''

        # Update the root to be for scratch
        self.root = Node('scratch', {'size': 0, 'Name': 'scratch'}, tag=None)
        self.data = None
        self._first_level = first_level

        # Count the nodes
        self.count = 1

        # Keep an index of all nodes
        self._index = {'scratch': ''}     

        # Sets self.data and builds self.tree
        if inputs != None:
            self.load(inputs)

        # If data is loaded, make the tree
        if self.data:
            self._make_tree(tag=tag, first_level=first_level)

    def __str__(self):
        return "CollectionTree<%s>" % self.count
    def __repr__(self):
        return "CollectionTree<%s>" % self.count


# Add / Remove Operations


    def remove(self, name, node=None, tag=None):
        '''find a path in the tree and remove (and return) the node if found.
           The function returns None if the Node wasn't in the tree (and wasn't
           found and removed). The way this is designed, we cannot remove the
           root node (but we just return it). If a tag is not provided,
           remove entire node with children tags.
         '''

        # Special case of first call, run function on root
        if node == None:
            node = self.root
            return self.remove(name, node, tag)

        # Set removed node to None
        removed = None

        # Iterate through children, and look for name
        # The top level of the tree doesn't have tags
        if isinstance(node.children, list):
            for child in node.children:

                # Did we find the node?
                if child.label == name:

                    # Case 1: we are given a tag
                    if tag != None:
                        if tag in child.children:
                            del child.children[tag]

                    # Case 2: we delete the entire node (and all tags)
                    else:
                        node.children = [c for c in node.children if c != child]

                    # Return the deleted node
                    self.count -= 1
                    return child
                
                # If we don't find the node, look for it in child
                removed = self.remove(name, child, tag)
                if removed != None:
                    return removed

        # All other nodes have dictionaries
        else:

            for child_tag, children in node.children.items():

                for child in children:

                    # Did we find the node?
                    if child.label == name:

                        # Case 1: we are given a tag
                        if tag != None:
                            if tag in child.children:
                                del child.children[tag]

                        # Case 2: we delete the entire node (and all tags) from list
                        else: 
                            childs = node.children[child_tag]
                            childs = [c for c in childs if c != child]
                            node.children[child_tag] = childs

                        # Return the deleted node
                        self.count -= 1
                        return child
                  
                    # If we don't find the node, look for it in child
                    removed = self.remove(name, child, tag)
                    if removed != None:
                        return removed

        return removed


# Searching Functions

    def find(self, name, tag=None, node=None):
        '''find a container in the tree and return the node if found.
           This base function is suited for searches that don't build
           on themselves (e.g., not filepaths or words). If the user
           specifies a tag, return the node only if the tag is included.
           For addition of a new tag to a node, leave the tag blank to
           return the general collection node (and then add it)
        '''
        # Only do index search if started from root
        if node == None:
            node = self.root
            found = self._find(name, tag, node)
            if found != None:
                return found

        # Did we find the node?
        if node.label == name:

            # if no tag is defined, return the entire node
            if tag == None:
                return node

            # Unless they want a tag, return the tag
            if tag in node.children:
                return node

            # Otherwise, we found the node, but not the right tag
            else:
                return None
              
        # Search through children
        for child in node.get_children():
            found_node = self.find(name, tag, child)
            if found_node != None:
                return found_node


    def _find(self, name, tag, node):
        '''a helper to find that tries finding with the index first.
        '''
        found = None

        # Otherwise, check the index for the node
        if name in self._index:
            lookup = self._index[name].split('|')

            # Pop the root node
            lookup.pop(0)
            children = self.root.children

            while lookup:
                childname = lookup.pop(0)
                childtag = lookup.pop(0)

                node = [x for x in children if x.name == childname][0]
                children = node.children[childtag]
 
            for found in children:
                if found.name == name:
                    return found               

        return found

    def search(self, name, number=None, node=None, tag=None, exact=False):
        '''find a basename in the tree. If number is defined, return
           up to that number. If exact is True, only grab exact matches
        '''
        found = []
 
        # We always start at the root  
        if node == None:
            node = self.root

        # Look for the name in the current node
        if name in node.label:
            
            if tag == None:

                # Only add exact matches
                if exact == True and name == node.label:
                    found.append(node)
                elif exact == False:
                    found.append(node)
            else:
                if tag in node.children:

                    # Only add exact matches
                    if exact == True and name == node.label:
                        found.append(node)

                    # Allow user to specify name via <collection>/<repo>:<tag>
                    elif exact == True and name == "%s:%s" % (node.label, tag):
                        found.append(node)

        # No children, no search
        if len(node.children) == 0:
            return found

        # Does the user want to cut out early?
        if number != None:
            if len(found) >= number:
                return found

        # Recursively search over children
        for child in node.get_children():
            found += self.search(name, number, child, tag, exact)
            if number != None:
                if len(found) >= number:
                    return found

        return found           


    def trace(self, name, node=None):
        '''find a path in the tree and return the node if found.
            We don't take a tag, assuming that the node returned is unique for
            the container name, and thus will have any associated tags.
         '''
        if node == None:
            node = self.root

        # Find the node, is it in the tree?
        tracedNode = self.find(name)

        # Trace it's path
        if tracedNode != None:
            return self._trace(name, node, tracedNode)


    def _trace(self, name, node, targetNode, traces=None):
        '''find a path in the tree and return the node if found.
        '''
        if traces == None:
            traces = []

        # Always add node we are searching
        traces.append(node)

        # Case 1: the node IS the target node
        if node == targetNode:
            return traces

        # Case 2: We found a leaf node, so we need to reset traces
        if node.leaf == True:
            return []

        else:
            for child in node.get_children():
                traces = self._trace(name, child, targetNode, traces)
                if targetNode in traces:
                    return traces
                traces = []

        return traces


# Loading Functions


    def load(self, uri, fromuri, tag=None):
        ''' uri must be the uri for a container. fromuri can be a uri OR
            a Dockerfile for the uri given (to extract from)
        '''
        self.update(uri, fromuri, tag)


    def _load(self, uri, fromuri):
        '''when we start here, we've been passed a Dockerfile.
        '''

        # Case 1. if we loaded a container-diff uri, it will be a uri
        if not "Dockerfile" in fromuri:     
            return {"Image": uri, "From": fromuri}

        # Case 2: Load the FROM in the Dockerfile            
        if os.path.exists(fromuri):
            froms = [x for x in self._load_dockerfile(fromuri, action="FROM") if x]

            # If we find the FROM, we want to extract history to get all from
            # In future, could support more than one from here
            if len(froms) > 0:
                 
                # We can't use any AS statements 
                if " as " in froms[0].lower():
                    return None

                # Remove any extra comments, etc. next to FROM <container>
                fromuri = froms[0].split(' ')[0]

                # Validate the uris
                for image in [uri, fromuri]:
                    if not parse_image_uri(image):
                        return None
                    # Don't allow uppercase or invalid endings
                    if image.isupper() or re.search('-$', image):
                        return None

                # Don't allow environment vars or similar
                literals = ['$', '}', '{', "'", '"']
                expression = "(%s)" % '|'.join(["[%s]" %x for x in literals])
                regexp = re.compile(expression)

                # Another condition for not parsing, if a variable is in uri
                if not regexp.search(uri) and not regexp.search(fromuri):
                    return {"Image": uri, "From": fromuri}


    def _load_dockerfile(self, dockerfile, action=None):
        '''extract one or more actions from a Dockerfile. If action is not
           defined, then return all that are found.
        '''
        lines = []
        with open(dockerfile, 'r') as filey:
            for line in filey.readlines():
                if action != None:
                    match = re.search("^%s" % action.upper(), line.strip().upper())
                    if match:
                        line = (line[match.end()+1:]).strip()
                        lines.append(line)
                else:
                    lines.append(line)
        return lines



    def update(self, uri, fromuri, tag=None):
        '''update will load in new data (without distributing an old self.data)
        '''
        # Loads {"Image": ... "From": ...}, unless Dockerfile not found
        data = self._load(uri, fromuri)

        # If we have loaded data, continue, returns True if nodes created/added
        if data:
            return self._make_tree(data['Image'], data['From'], tag=tag)


# Tree Generation


    def _make_tree(self, uri, fromuri, tag=None):
        '''construct the tree from the loaded data (self.data) which for
           this collection tree is a URI for a Docker image. Since we are making
           the tree starting with a single container and building based on
           FROM statements, the nodes correspond to different containers,
           and they are tagged with container tags.

           The tree has a structure that has namespace as node, and tags
           as internal dictionaries to point to lists of children. For
           example:
 
               Node<mhart/alpine-node>
                  {"children":     
                      {"4": [] }
                      {"6": [] }
                  }

           When we export the tree, the children get combined into one list
           and tagged with their parent tag.

           Parameters
           ==========
           uri: the container uri to add
           fromuri: the from uri (the parent of the container, or it's base)
           tag: if defined, a tag to give the node (not the container tag)

           Returns
           =======
           created: boolean to indicate if the nodes were created and/or added.
                    This boolean should be used to keep track of pairs that
                    weren't able to be added to the tree because of not having
                    a connection to the root.
        '''
        added = False

        # Parse the uri to get namespace, and rest
        uriImage = parse_image_uri(uri)
        uriFrom = parse_image_uri(fromuri)        

        # Update uri and fromuri
        uri = uriImage['nodeUri']

        # Special case of scratch
        if fromuri != 'scratch':
            fromuri = uriFrom['nodeUri']

        # Find the node in the tree, if it exists (the uri without tag/version)
        nodeImage = self.find(uri)
        nodeFrom = self.find(fromuri) 

        # If they both exist, do nothing.
        if nodeImage != None and nodeFrom != None:
            nodeImage.counter +=1
            return True

        # Boolean to indicate add to root
        append_root = False
 
        # If FROM not in tree, create
        if nodeFrom == None:
            nodeFrom = MultiNode(fromuri, {"Name": fromuri })
            append_root = True

        # If the tag isn't there, add it (this is the parent)
        if isinstance(nodeFrom.children, dict):
            if uriFrom['repo_tag'] not in nodeFrom.children:
                nodeFrom.children[uriFrom['repo_tag']] = []

        # If node Image isn't in tree, create it
        if nodeImage == None:
            nodeImage = MultiNode(uri, {"Name": uri })
            nodeImage.leaf = True
            self.count += 1
        else:
            nodeImage.counter += 1

        # Add the (child) tag if it's not there
        if uriImage['repo_tag'] not in nodeImage.children:
            nodeImage.children[uriImage['repo_tag']] = []

        # Remove the nodeImage from the tree, if it's found
        self.remove(name = nodeImage.label)

        # Add the tag, if defined (this is not the uri tag, but another)
        if tag is not None:
            if tag not in nodeImage.tags:
                nodeImage.tags.add(tag)
       
        # Special case for the root node, update the lookup and add first level
        if nodeFrom == self.root:
            self._index[nodeImage.name] = self._index[nodeFrom.name]

            # The image isn't in root's children, but we can add it there!
            if nodeImage not in nodeFrom.children and nodeImage.label.startswith(self._first_level):
                nodeFrom.children.append(nodeImage)    
                added = True

        else:

            # We now have a nodeFrom and a nodeImage, we can append        
            if nodeImage not in nodeFrom.children[uriFrom['repo_tag']]:
                nodeFrom.children[uriFrom['repo_tag']].append(nodeImage)

            # We only append library to the root.
            if append_root is True and nodeFrom.label.startswith(self._first_level):

                # If adding to the root, we likely don't have the parent node
                if nodeFrom.name not in self._index:
                    self._index[nodeFrom.name] = self._index[self.root.name]

                self._index[nodeImage.name] = '%s|%s|%s' %(self._index[nodeFrom.name],
                                                           nodeFrom.name,
                                                           uriFrom['repo_tag'])

                if nodeFrom not in self.root.children:
                    self.root.children.append(nodeFrom)
                
                added = True

        # If it was a leaf, no longer is
        nodeFrom.leaf = False
        return added


# Export Functions

    def paths(self, leaves_only=False, 
                    add_tags=True, 
                    tag_prefix='.',
                    label=None):

        '''Get all paths to nodes, as an iterator

           Parameters
           ==========
           leaves_only: only get leaf nodes
           add_tags: add tags to tree (also as folders)
           tag_prefix: if adding tags, use this prefix.
           label: get the path for one (string) or more (list) nodes.
        '''

        def traverse(current, path='', label=None):

            # Update the path with the current
            path = path + '/' + current.name

            # Does the user want to export leaves only?
            if (leaves_only and current.leaf) or not leaves_only:

                # Does the user only want specific label(s)?
                if label != None:
                    if current.name in label:
                        label.remove(current.name)
                        yield path
                else:
                    yield path

            # Case 1: We are at the root (and have list)
            if isinstance(current.children, list):
                for child in current.children:
                    for new_path in traverse(child, path, label):
                        yield new_path

            # Case 2: we have a dictionary
            else:
                for tag, children in current.children.items():
                    for child in children:

                        # tags are represented by hidden folders
                        if add_tags:
                            new_path = path + '/' + tag_prefix + tag

                        # But if the user doesn't want that detail
                        else:
                            new_path = path

                        # Reveal the paths
                        for new_path in traverse(child, new_path, label):
                            yield new_path

        # Does the user want a particular set of labels?
        if label != None:
            if not isinstance(label, list):
                label = [label]

            # Create a set for more efficient lookup
            label = set(label)

        for path in traverse(self.root, label=label):
            yield path


    def get_paths(self, leaves_only=False, add_tags=True, tag_prefix='.', label=None):

        '''Get all paths to nodes, in a list.

           Parameters
           ==========
           leaves_only: only get leaf nodes
           add_tags: add tags to tree (also as folders)
           tag_prefix: if adding tags, use this prefix.
           label: get the path for one (string) or more (list) nodes.
        '''
        return list(self.paths(add_tags=add_tags,
                               leaves_only=leaves_only, 
                               tag_prefix=tag_prefix,
                               label=label))


    def get_nodes(self):
        '''return a list of nodes (a call to __iter__)
        '''
        return list(self.__iter__())


    def __iter__(self):
        '''an iterator over MultiNodes to yield nodes, one at a time.
        '''
        def traverse(current, seen):
            for child in current.get_children():
                if child.name not in seen:
                    seen.add(child.name)
                    yield child
                    for child in traverse(child, seen):
                        yield child

        seen = set()
        for node in traverse(self.root, seen):
            yield node  


    def export_tree(self, filename=None):
        '''export a data structure for a weighted, colored tree, either just
           the data or the full html / visualization. If filename is defined,
           export the raw data json to the file. If the generate_html is
           True, do the same but embed in html.

           Parameters
           ==========
           filename: if defined, write data to file (and return) otherwise
           return data structure

        '''

        # We will call this recursively
 
        def traverse(nodes={}, current=None):

            colors = ['#0000FF', # blue
                      '#FF7F00', # orange
                      '#FF0000', # red
                      '#7F007F', # purple
                      '#00FFFF', # cyan
                      '#0560D0'] # generic blue

            if current is None:
                current = self.root

            tags = list(current.tags)

            # If not the root node (scratch, tags are in keys)
            if not isinstance(current.children, list):
                tags += list(current.children.keys())

            new_node = {'color': choice(colors),
                        'key': current.label,
                        'name': current.label,
                        'tags': tags,
                        'attrs': current.get_attributes(),
                        'children': [] }

            # Add the size if was provided!
            if hasattr(current, 'size'):
                new_node['size'] = current.size

            if len(nodes) == 0:
                nodes.update(new_node)
            else:            
                nodes['children'].append(new_node)

            # Traverse remainder of children
            for child in current.get_children():
                traverse(nodes=new_node, current=child)

        nodes = dict()
        traverse(nodes)
                
        # If the user provided a file, export to it
        if filename is not None:
            with open(filename, 'w') as filey:
                filey.writelines(json.dumps(nodes))
            return filename

        return nodes

       
    def get_count(self, name, tag=None):
        '''find a path in the tree and return the node if found
        '''
        counter = 0
        node = self.find(name, tag=tag)
        if node:
            counter = node.counter
        return counter


# Loading Functions
CollectionTree._update = _update
CollectionTree._load_http = _load_http
CollectionTree._load_list = _load_list
CollectionTree._load_json = _load_json
CollectionTree._load_container_diff = _load_container_diff



#!/usr/bin/env python

#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import containertree
import argparse
import sys
import os


def get_parser():
    parser = argparse.ArgumentParser(description="ContainerTrees in Python")

    # Global Variables
    parser.add_argument('--debug', dest="debug", 
                        help="use verbose logging to debug.", 
                        default=False, action='store_true')

    parser.add_argument('--quiet', dest="quiet", 
                        help="suppress additional output.", 
                        default=False, action='store_true')

    parser.add_argument('--version', dest="version", 
                        help="print version and exit.", 
                        default=False, action='store_true')

    description = 'actions for ContainerTree Python'
    subparsers = parser.add_subparsers(help='containertree actions',
                                       title='actions',
                                       description=description,
                                       dest="command")

    # View templates
    templates = subparsers.add_parser("templates",
                                       help="View available tree templates")

    # Generate Tree
    generate = subparsers.add_parser("generate",
                                      help="Generate a container tree.")

    generate.add_argument("image", nargs=1,
                          help='Docker Container URI', 
                          type=str)

    generate.add_argument('--force','-f', dest="force", 
                          help="force generate if output file exists.", 
                          default=False, action='store_true')

    generate.add_argument('--view', dest="view", 
                           help="open a web browser to view tree.", 
                           default=False, action='store_true')

    generate.add_argument('--template', dest="template", 
                           help="templates to choose from for tree", 
                           default='files_tree', type=str)

    generate.add_argument('--output', dest="output", 
                           help="output folder to generate static files", 
                           default=None, type=str)

    generate.add_argument('--print', dest="printout", 
                           help="print a specific output file to the terminal", 
                           default=None, type=str)

    return parser


def get_subparsers(parser):
    '''get_subparser will get a dictionary of subparsers, to help with printing help
    '''

    actions = [action for action in parser._actions 
               if isinstance(action, argparse._SubParsersAction)]

    subparsers = dict()
    for action in actions:
        # get all subparsers and print help
        for choice, subparser in action.choices.items():
            subparsers[choice] = subparser

    return subparsers



def main():
    '''allow the user to generate container trees from the command line
    '''

    parser = get_parser()
    subparsers = get_subparsers(parser)

    def help(return_code=0):
        '''print help, including the software version and active client 
           and exit with return code.
        '''

        version = containertree.__version__

        print("\nContainerTree v%s" % version)
        parser.print_help()
        sys.exit(return_code)
    
    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()
    try:
        args = parser.parse_args()
    except:
        sys.exit(0)

    if args.quiet is True:
        os.environ['MESSAGELEVEL'] = "QUIET"
    elif args.debug is False:
        os.environ['MESSAGELEVEL'] = "DEBUG"

    # Show the version and exit
    if args.version:
        print(containertree.__version__)
        sys.exit(0)

    # Does the user want a shell?
    if args.command == "generate": from .generate import main
    elif args.command == "templates": from .templates import main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)

if __name__ == '__main__':
    main()



#
# Copyright (C) 2018-2019 Vanessa Sochat.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .collection import CollectionTree
from .container import ( 
    ContainerTree, 
    ContainerFileTree,
    ContainerAptTree,
    ContainerPipTree
)



'''

Copyright (C) 2018-2019 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

import json
import math
import os
import re
import requests
import shutil
import sys
import tempfile
from .https import get, call

########################################################################
# Parsing URIs: we assume the user will provide registry and/or ports
###############################################################################

_reduced_uri = re.compile("(?:(?P<registry>[^/@]+[.:][^/@]*)/)?"
                          "(?P<repo>[^:@/]+)"
                          "(?::(?P<tag>[^:@]+))?"
                          "(?:@(?P<version>.+))?"
                          "$"
                          "(?P<namespace>.)?")

_docker_uri = re.compile("(?:(?P<registry>[^/@]+[.:][^/@]*)/)?"
                         "(?P<namespace>(?:[^:@/]+/)+)?"
                         "(?P<repo>[^:@/]+)"
                         "(?::(?P<tag>[^:@]+))?"
                         "(?:@(?P<version>.+))?"
                         "$")

_default_uri = re.compile("(?:(?P<registry>[^/@]+)/)?"
                          "(?P<namespace>(?:[^:@/]+/)+)"
                          "(?P<repo>[^:@/]+)"
                          "(?::(?P<tag>[^:@]+))?"
                          "(?:@(?P<version>.+))?"
                          "$")

def parse_image_uri(image, default_tag='latest', default_namespace='library'):
    '''parse the image uri and return a dictionary to look up namespace,
       tag, and registry. If the name doesn't match a pattern, we return None.

       Parameters
       ==========
       image: the full image uri (e.g., library/ubuntu:latest
       default_tag: the image tag (e.g., latest) default is latest
       default_namespace: the default collection (e.g., library)
    '''

    # Match from docker to default
    uri_regexes = [ _docker_uri,
                    _reduced_uri,
                    _default_uri ]

    for r in uri_regexes:
        match = r.match(image)
        if match:
            break

    if not match:
        # Calling client should expect None to indicate not parseable
        print('Could not parse image %s' % image)
        return None

    registry = match.group('registry')
    namespace = match.group('namespace')
    repo_name = match.group('repo')
    repo_tag = match.group('tag')
    version = match.group('version')

    if namespace:
        namespace = namespace.rstrip('/')

    # replace empty fields with defaults
    if not namespace:
        namespace = default_namespace
    if not repo_tag:
        repo_tag = default_tag

    # Full uri (without registry)
    nodeuri = "%s/%s" %(namespace, repo_name)
    fulluri = nodeuri
    if repo_tag != None:
        fulluri = "%s:%s" %(fulluri, repo_tag)
    if version != None:
        fulluri = "%s@%s" %(fulluri, version)

    # We don't require registry, or version.
    parsed = {'registry': registry,
              'namespace': namespace,
              'repo_name': repo_name,
              'repo_tag': repo_tag,
              'version': version,
              'fullUri': fulluri,
              'nodeUri': nodeuri }

    return parsed


class DockerInspector(object):

    def __init__(self, container_name=None, base=None, version=None):
        self.manifests = {}
        self.uri = container_name
        self._set_base(base, version)
        self.headers = {'Content-Type':"application/json"}

    def __str__(self):
        return "DockerInspector<%s>" % self.uri
    def __repr__(self):
        return "DockerInspector<%s>" % self.uri


    def _set_base(self, base=None, version=None):
        '''set the API base or default to use Docker Hub. 
        '''
        if base is None:
            base = "index.docker.io"

        if version is None:
            version = "v2"

        # <protocol>://<base>/<version>

        self._base = "https://%s" % base
        self._version = version
        self.base = "%s/%s" %(self._base.strip('/'), version)
    
    def update_token(self, response):
        '''update_token uses HTTP basic authentication to get a token for
           Docker registry API V2 operations. We get here if a 401 is
           returned for a request.
    
           Parameters
           ==========
           response: the http request response to parse for the challenge.
    
           https://docs.docker.com/registry/spec/auth/token/
        '''

        not_asking_auth = "Www-Authenticate" not in response.headers
        if response.status_code != 401 or not_asking_auth:
            print("Authentication error, exiting.")
            sys.exit(1)

        challenge = response.headers["Www-Authenticate"]
        regexp = '^Bearer\s+realm="(.+)",service="(.+)",scope="(.+)",?'
        match = re.match(regexp, challenge)

        if not match:
            print("Unrecognized authentication challenge, exiting.")
            sys.exit(1)

        realm = match.group(1)
        service = match.group(2)
        scope = match.group(3).split(',')[0]

        token_url = realm + '?service=' + service + '&expires_in=900&scope=' + scope
    
        # Default headers must be False so that client's current headers not used
        response = self._get(token_url)

        try:
            token = response["token"]
            token = {"Authorization": "Bearer %s" % token}
            self.headers.update(token)

        except Exception:
            print("Error getting token.")
            sys.exit(1)

    def get_manifests(self, digest=None):
        '''get_manifests calls get_manifest for each of the schema versions,
           including v2 and v1. Version 1 includes image layers and metadata,
           and version 2 must be parsed for a specific manifest, and the 2nd
           call includes the layers. If a digest is not provided
           latest is used.

           Parameters
           ==========
           repo_name: reference to the <username>/<repository>:<tag> to obtain
           digest: a tag or shasum version
        '''
    
        # Obtain schema version 1 (metadata) and 2, and image config
        schemaVersions = ['v1', 'v2', 'config']
        for schemaVersion in schemaVersions:
            manifest = self._get_manifest(digest, schemaVersion)
            if manifest is not None:

                # If we don't have a config yet, try to get from version 2 manifest
                if schemaVersion == "v2" and "config" in manifest:
                    print('Attempting to get config as blob in verison 2 manifest')
                    url = self._get_layerLink(repo_name, manifest['config']['digest'])        
                    headers = {'Accept': manifest['config']['mediaType']}
                    self.manifests['config'] = self._get(url, headers=headers)
    
                self.manifests[schemaVersion] = manifest

        return self.manifests


    def get_manifest_selfLink(self, digest=None):
        ''' get a selfLink for the manifest, for use by the client get_manifest
            function, along with the parents pull
     
           Parameters
           ==========
           digest: a tag or shasum version

        '''
        url = "%s/%s/manifests" % (self.base, self.uri)

        # Add a digest - a tag or hash (version)
        if digest is None:
            digest = 'latest'
        return "%s/%s" % (url, digest)


    def get_manifest(self, digest=None, version="v1"):
        '''
           get_manifest should return an image manifest
           for a particular repo and tag.  The image details
           are extracted when the client is generated.
    
           Parameters
           ==========
           digest: a tag or shasum version
           version: one of v1, v2, and config (for image config)

        '''
        manifest = None
        accepts = {'config': "application/vnd.docker.container.image.v1+json",
                   'v1': "application/vnd.docker.distribution.manifest.v1+json",
                   'v2': "application/vnd.docker.distribution.manifest.v2+json" }

        url = self.get_manifest_selfLink(digest)
    
        print("Obtaining manifest: %s" % url)
        headers = {'Accept': accepts[version] }

        try:
            manifest = self._get(url, headers=headers)
            manifest['selfLink'] = url
        except:
            pass

        return manifest
 

################################################################################
# Metadata (Environment, Labels, Runscript)
################################################################################


    def get_config(self, key="Entrypoint", delim=None):
        '''get_config returns a particular key (default is Entrypoint)
            from a VERSION 1 manifest obtained with get_manifest.
    
            Parameters
            ==========
            key: the key to return from the manifest config
            delim: Given a list, the delim to use to join the entries.
            Default is newline

        '''
        cmd = None
    
        # If we didn't find the config value in version 2
        for version in ['config', 'v1']:
            if cmd is None and 'config' in self.manifests:
                
                # First try, version 2.0 manifest config has upper level config
                manifest = self.manifests['config']
                if "config" in manifest:
                    if key in manifest['config']:
                        cmd = manifest['config'][key]
    
                # Second try, config manifest (not from verison 2.0 schema blob)
    
                if cmd is None and "history" in manifest:
                    for entry in manifest['history']:
                        if 'v1Compatibility' in entry:
                            entry = json.loads(entry['v1Compatibility'])
                            if "config" in entry:
                                if key in entry["config"]:
                                    cmd = entry["config"][key]
    
        # Standard is to include commands like ['/bin/sh']
        if isinstance(cmd, list):
            if delim is not None:
                cmd = delim.join(cmd)
        print("Found Docker config (%s) %s" % (key, cmd))
        return cmd

DockerInspector._get = get
DockerInspector._call = call
